/*
 * Copyright 2001-2005, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <sys/sysmacros.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <error.h>
#include <sys/ioctl.h>
#include <sys/sysinfo.h>

#include <linux/mtd/mtd.h>

#include <trxhdr.h>
#include <rts/crc.h>
#include <bcmutils.h>
#include <shutils.h>
#include <bcmnvram.h>

#define MOTOROLA_SERIES_CHECK_FIRMWARE
#if defined(MOTOROLA_SERIES_CHECK_FIRMWARE)

static char VDID[4] = {	0x10, 0x57, 0x70, 0x00};
#endif
#define ROUNDUP(x, y) ((((unsigned int)(x)+((y)-1))/(y))*(y))
#define OFFSET(p, field) ((unsigned int)(&((p)->field)) - (unsigned int)(p))
#define MIN(a, b) \
({ typedef _ta = (a), _tb = (b); \
   _ta _a = (a); _tb _b = (b); \
   _a < _b ? _a : _b; \
})

/* crc defines */
#define CRC32_INIT_VALUE 0xffffffff /* Initial CRC32 checksum value */
#define CRC32_GOOD_VALUE 0xdebb20e3 /* Good final CRC32 checksum value */

/*******************************************************************************
 * crc32
 *
 * Computes a crc32 over the input data using the polynomial:
 *
 *       x^32+x^26+x^23+x^22+x^16+x^12+x^11+X^10+x^8+x^7+x^5+x^4+x^2+x+1
 *
 * The caller provides the initial value (either CRC32_INIT_VALUE
 * or the previous returned value) to allow for processing of 
 * discontiguous blocks of data.  When generating the CRC the
 * caller is responsible for complementing the final return value
 * and inserting it into the byte stream.  When checking, a final
 * return value of CRC32_GOOD_VALUE indicates a valid CRC.
 *
 * Reference: Dallas Semiconductor Application Note 27
 *   Williams, Ross N., "A Painless Guide to CRC Error Detection Algorithms", 
 *     ver 3, Aug 1993, ross@guest.adelaide.edu.au, Rocksoft Pty Ltd.,
 *     ftp://ftp.rocksoft.com/clients/rocksoft/papers/crc_v3.txt
 *
 ******************************************************************************/

static u_int32_t crc32_table[256] = {
    0x00000000, 0x77073096, 0xEE0E612C, 0x990951BA,
    0x076DC419, 0x706AF48F, 0xE963A535, 0x9E6495A3,
    0x0EDB8832, 0x79DCB8A4, 0xE0D5E91E, 0x97D2D988,
    0x09B64C2B, 0x7EB17CBD, 0xE7B82D07, 0x90BF1D91,
    0x1DB71064, 0x6AB020F2, 0xF3B97148, 0x84BE41DE,
    0x1ADAD47D, 0x6DDDE4EB, 0xF4D4B551, 0x83D385C7,
    0x136C9856, 0x646BA8C0, 0xFD62F97A, 0x8A65C9EC,
    0x14015C4F, 0x63066CD9, 0xFA0F3D63, 0x8D080DF5,
    0x3B6E20C8, 0x4C69105E, 0xD56041E4, 0xA2677172,
    0x3C03E4D1, 0x4B04D447, 0xD20D85FD, 0xA50AB56B,
    0x35B5A8FA, 0x42B2986C, 0xDBBBC9D6, 0xACBCF940,
    0x32D86CE3, 0x45DF5C75, 0xDCD60DCF, 0xABD13D59,
    0x26D930AC, 0x51DE003A, 0xC8D75180, 0xBFD06116,
    0x21B4F4B5, 0x56B3C423, 0xCFBA9599, 0xB8BDA50F,
    0x2802B89E, 0x5F058808, 0xC60CD9B2, 0xB10BE924,
    0x2F6F7C87, 0x58684C11, 0xC1611DAB, 0xB6662D3D,
    0x76DC4190, 0x01DB7106, 0x98D220BC, 0xEFD5102A,
    0x71B18589, 0x06B6B51F, 0x9FBFE4A5, 0xE8B8D433,
    0x7807C9A2, 0x0F00F934, 0x9609A88E, 0xE10E9818,
    0x7F6A0DBB, 0x086D3D2D, 0x91646C97, 0xE6635C01,
    0x6B6B51F4, 0x1C6C6162, 0x856530D8, 0xF262004E,
    0x6C0695ED, 0x1B01A57B, 0x8208F4C1, 0xF50FC457,
    0x65B0D9C6, 0x12B7E950, 0x8BBEB8EA, 0xFCB9887C,
    0x62DD1DDF, 0x15DA2D49, 0x8CD37CF3, 0xFBD44C65,
    0x4DB26158, 0x3AB551CE, 0xA3BC0074, 0xD4BB30E2,
    0x4ADFA541, 0x3DD895D7, 0xA4D1C46D, 0xD3D6F4FB,
    0x4369E96A, 0x346ED9FC, 0xAD678846, 0xDA60B8D0,
    0x44042D73, 0x33031DE5, 0xAA0A4C5F, 0xDD0D7CC9,
    0x5005713C, 0x270241AA, 0xBE0B1010, 0xC90C2086,
    0x5768B525, 0x206F85B3, 0xB966D409, 0xCE61E49F,
    0x5EDEF90E, 0x29D9C998, 0xB0D09822, 0xC7D7A8B4,
    0x59B33D17, 0x2EB40D81, 0xB7BD5C3B, 0xC0BA6CAD,
    0xEDB88320, 0x9ABFB3B6, 0x03B6E20C, 0x74B1D29A,
    0xEAD54739, 0x9DD277AF, 0x04DB2615, 0x73DC1683,
    0xE3630B12, 0x94643B84, 0x0D6D6A3E, 0x7A6A5AA8,
    0xE40ECF0B, 0x9309FF9D, 0x0A00AE27, 0x7D079EB1,
    0xF00F9344, 0x8708A3D2, 0x1E01F268, 0x6906C2FE,
    0xF762575D, 0x806567CB, 0x196C3671, 0x6E6B06E7,
    0xFED41B76, 0x89D32BE0, 0x10DA7A5A, 0x67DD4ACC,
    0xF9B9DF6F, 0x8EBEEFF9, 0x17B7BE43, 0x60B08ED5,
    0xD6D6A3E8, 0xA1D1937E, 0x38D8C2C4, 0x4FDFF252,
    0xD1BB67F1, 0xA6BC5767, 0x3FB506DD, 0x48B2364B,
    0xD80D2BDA, 0xAF0A1B4C, 0x36034AF6, 0x41047A60,
    0xDF60EFC3, 0xA867DF55, 0x316E8EEF, 0x4669BE79,
    0xCB61B38C, 0xBC66831A, 0x256FD2A0, 0x5268E236,
    0xCC0C7795, 0xBB0B4703, 0x220216B9, 0x5505262F,
    0xC5BA3BBE, 0xB2BD0B28, 0x2BB45A92, 0x5CB36A04,
    0xC2D7FFA7, 0xB5D0CF31, 0x2CD99E8B, 0x5BDEAE1D,
    0x9B64C2B0, 0xEC63F226, 0x756AA39C, 0x026D930A,
    0x9C0906A9, 0xEB0E363F, 0x72076785, 0x05005713,
    0x95BF4A82, 0xE2B87A14, 0x7BB12BAE, 0x0CB61B38,
    0x92D28E9B, 0xE5D5BE0D, 0x7CDCEFB7, 0x0BDBDF21,
    0x86D3D2D4, 0xF1D4E242, 0x68DDB3F8, 0x1FDA836E,
    0x81BE16CD, 0xF6B9265B, 0x6FB077E1, 0x18B74777,
    0x88085AE6, 0xFF0F6A70, 0x66063BCA, 0x11010B5C,
    0x8F659EFF, 0xF862AE69, 0x616BFFD3, 0x166CCF45,
    0xA00AE278, 0xD70DD2EE, 0x4E048354, 0x3903B3C2,
    0xA7672661, 0xD06016F7, 0x4969474D, 0x3E6E77DB,
    0xAED16A4A, 0xD9D65ADC, 0x40DF0B66, 0x37D83BF0,
    0xA9BCAE53, 0xDEBB9EC5, 0x47B2CF7F, 0x30B5FFE9,
    0xBDBDF21C, 0xCABAC28A, 0x53B39330, 0x24B4A3A6,
    0xBAD03605, 0xCDD70693, 0x54DE5729, 0x23D967BF,
    0xB3667A2E, 0xC4614AB8, 0x5D681B02, 0x2A6F2B94,
    0xB40BBE37, 0xC30C8EA1, 0x5A05DF1B, 0x2D02EF8D
};

#define CRC_INNER_LOOP(n, c, x) \
    (c) = ((c) >> 8) ^ crc##n##_table[((c) ^ (x)) & 0xff]

u_int32_t
crc32(
    u_int8_t *pdata,  /* pointer to array of data to process */
    unsigned int nbytes, /* number of input data bytes to process */
    u_int32_t crc     /* either CRC32_INIT_VALUE or previous return value */
)
{
    while (nbytes-- > 0)
        CRC_INNER_LOOP(32, crc, *pdata++);
    return crc;
}

/*
 * Open an MTD device
 * @param	mtd	path to or partition name of MTD device
 * @param	flags	open() flags
 * @return	return value of open()
 */
int
mtd_open(const char *mtd, int flags)
{
	FILE *fp;
	char dev[PATH_MAX];
	int i;

	if ((fp = fopen("/proc/mtd", "r"))) {
		while (fgets(dev, sizeof(dev), fp)) {
			if (sscanf(dev, "mtd%d:", &i) && strstr(dev, mtd)) {
				snprintf(dev, sizeof(dev), "/dev/mtd/%d", i);
				fclose(fp);
				return open(dev, flags);
			}
		}
		fclose(fp);
	}

	return open(mtd, flags);
}	

/*
 * Erase an MTD device
 * @param	mtd	path to or partition name of MTD device
 * @return	0 on success and errno on failure
 */
int
mtd_erase(const char *mtd)
{
	int mtd_fd;
	mtd_info_t mtd_info;
	erase_info_t erase_info;

	/* Open MTD device */
	if ((mtd_fd = mtd_open(mtd, O_RDWR)) < 0) {
		perror(mtd);
		return errno;
	}

	/* Get sector size */
	if (ioctl(mtd_fd, MEMGETINFO, &mtd_info) != 0) {
		perror(mtd);
		close(mtd_fd);
		return errno;
	}

	erase_info.length = mtd_info.erasesize;

	for (erase_info.start = 0;
	     erase_info.start < mtd_info.size;
	     erase_info.start += mtd_info.erasesize) {
		if (ioctl(mtd_fd, MEMERASE, &erase_info) != 0) {
			perror(mtd);
			close(mtd_fd);
			return errno;
		}
	}

	close(mtd_fd);
	return 0;
}

/* New mtd_write for RC 3.31.12.0 */
extern int http_get(const char *server, char *buf, size_t count, off_t offset);

/*
 * Write a file to an MTD device
 * @param	path	file to write or a URL
 * @param	mtd	path to or partition name of MTD device 
 * @return	0 on success and errno on failure
 */
int
mtd_write(const char *path, const char *mtd)
{
	int mtd_fd = -1;
	mtd_info_t mtd_info;
	erase_info_t erase_info;

	struct sysinfo info;
	struct trx_header trx;
	unsigned long crc;

	FILE *fp;
	char *buf = NULL;
	long count, len, off;
	int ret = -1;
#if defined(MOTOROLA_SERIES_CHECK_FIRMWARE)
        unsigned char t[8];
#endif
	/* Examine TRX header */
	if ((fp = fopen(path, "r")))
  {
#if defined(MOTOROLA_SERIES_CHECK_FIRMWARE)

		unsigned char moto_4702_fw[7][8] ={
		{0xE2, 0xD9, 0xD0, 0x5D, 0x10, 0x57, 0x70, 0x00}, //v2.00
		{0x87, 0xFD, 0xFC, 0x9A, 0x10, 0x57, 0x70, 0x00}, //v2.03
		{0xF6, 0xE1, 0x0E, 0x74, 0x10, 0x57, 0x70, 0x00}, //v2.04
		{0xB1, 0x26, 0xEB, 0xB4, 0x10, 0x57, 0x70, 0x00}, //v2.06
		{0xB8, 0xD8, 0xFC, 0x4C, 0x10, 0x57, 0x70, 0x00}, //v2.07
		{0xD7, 0xA9, 0x78, 0x48, 0x10, 0x57, 0x70, 0x00}, //v2.08
		{0x53, 0x69, 0xBA, 0x1F, 0x10, 0x57, 0x70, 0x00}  //v2.09
		};
		
		safe_fread(t, 1, 8, fp);
		
		if(!strcmp(nvram_safe_get("cpu_type"), "BCM3302") || !strcmp(nvram_safe_get("cpu_type"), "BCM4712"))
		{
			if(!memcmp(t, moto_4702_fw[0], 8) || !memcmp(t, moto_4702_fw[1], 8) || !memcmp(t, moto_4702_fw[2], 8) || !memcmp(t, moto_4702_fw[3], 8) || !memcmp(t, moto_4702_fw[4], 8) || !memcmp(t, moto_4702_fw[5], 8) || !memcmp(t, moto_4702_fw[6], 8)) 
			{
				fprintf(stderr, "4702 only firmware...operate fail on 4712\n");
				ret = 4702;
				goto fail;
			}
		}
		
		if(memcmp( t + 4, VDID, 4 ))
		{
			//fprintf(stderr, "Model Number not match.\n" );
			goto fail ;
		}
		else
		{
			//fprintf(stderr, "Model Number right.\n" );
		}
		count = safe_fread(&trx, 1, sizeof(struct trx_header), fp);
#else
		count = safe_fread(&trx, 1, sizeof(struct trx_header), fp);
#endif
        }
	else
		count = http_get(path, (char *) &trx, sizeof(struct trx_header), 0);
	if (count < sizeof(struct trx_header)) {
		fprintf(stderr, "%s: File is too small (%ld bytes)\n", path, count);
		goto fail;
	}
	if (trx.magic != TRX_MAGIC ||
	    trx.len > TRX_MAX_LEN ||
	    trx.len < sizeof(struct trx_header)) {
		fprintf(stderr, "%s: Bad trx header\n", path);
		goto fail;
	}

	/* Open MTD device and get sector size */
	if ((mtd_fd = mtd_open(mtd, O_RDWR)) < 0 ||
	    ioctl(mtd_fd, MEMGETINFO, &mtd_info) != 0 ||
	    mtd_info.erasesize < sizeof(struct trx_header)) {
		perror(mtd);
		goto fail;
	}

	/* See if we have enough memory to store the whole file */
	sysinfo(&info);
	if ((info.freeram + info.bufferram) >= trx.len)
		erase_info.length = ROUNDUP(trx.len, mtd_info.erasesize);
	else
		erase_info.length = mtd_info.erasesize;

	/* Allocate temporary buffer */
//	fprintf(stderr, "mtd : allocate memory address size %d\n", erase_info.length);
	
	if (!(buf = malloc(erase_info.length))) {
		perror("malloc");
		goto fail;
	}

	/* Calculate CRC over header */
	crc = crc32((uint8 *) &trx.flag_version, sizeof(struct trx_header) - OFFSETOF(struct trx_header, flag_version), CRC32_INIT_VALUE);

	if (trx.flag_version & TRX_NO_HEADER)
		trx.len -= sizeof(struct trx_header);

	/* Write file or URL to MTD device */
	for (erase_info.start = 0; erase_info.start < trx.len; erase_info.start += count) {
		len = MIN(erase_info.length, trx.len - erase_info.start);
		if ((trx.flag_version & TRX_NO_HEADER) || erase_info.start)
			count = off = 0;
		else {
			count = off = sizeof(struct trx_header);
			memcpy(buf, &trx, sizeof(struct trx_header));
		}
		if (fp)
		{
      count += safe_fread(&buf[off], 1, len - off, fp);
#if defined(MOTOROLA_SERIES_CHECK_FIRMWARE)
			if( !strcmp( mtd, "linux" ) )
			{
				//printf("buf : %02X %02X %02X %02X\n", buf[0], buf[1], buf[2], buf[3]);
			}
#endif
		}
		else
			count += http_get(path, &buf[off], len - off, erase_info.start + off);
		if (count < len) {
			fprintf(stderr, "%s: Truncated file (actual %ld expect %ld)\n", path,
				count - off, len - off);
			goto fail;
		}
		/* Update CRC */
		crc = crc32(&buf[off], count - off, crc);
		/* Check CRC before writing if possible */
		if ((count) == trx.len) {
			if (crc != trx.crc32) {
				//fprintf(stderr, "%s: Bad CRC\n", path);
				goto fail;
			} 
			else
			{
				//fprintf(stderr, "CRC correct...\n");
			}
#if defined(MOTOROLA_SERIES_CHECK_FIRMWARE)
      unsigned int moto_crc = 0;
			moto_crc = crc32(t + 4, 2, CRC32_INIT_VALUE);
			moto_crc = crc32(t + 6, 2, moto_crc);
			moto_crc = crc32(buf, trx.len, moto_crc);

			if(!memcmp(t + 3, (char *)&moto_crc, 1) && !memcmp(t + 2, (char *)&moto_crc + 1, 1) && !memcmp(t + 1, (char *)&moto_crc + 2, 1) && !memcmp(t + 0, (char *)&moto_crc + 3, 1))
			{
				//fprintf(stderr, "Motorola CRC correct...\n");
			}
			else
			{
				//fprintf(stderr, "Bad Motorola CRC\n");
				goto fail;
			}
#endif
		}
		
		/* Do it */
		(void) ioctl(mtd_fd, MEMUNLOCK, &erase_info);
		if (ioctl(mtd_fd, MEMERASE, &erase_info) != 0 || write(mtd_fd, buf, count) != count) 
		{
		perror(mtd);
		goto fail;
		}
	}

	//fprintf(stderr, "%s: CRC OK\n", mtd);
	ret = 0;

 fail:
	if (buf) {
		/* Dummy read to ensure chip(s) are out of lock/suspend state */
		(void) read(mtd_fd, buf, 2);
		free(buf);
	}

	if (mtd_fd >= 0)
		close(mtd_fd);
	if (fp)
		fclose(fp);
	return ret;
}

/* Old mtd_write for RC 3.11.48.7 */
//extern int http_get(const char *server, char *buf, size_t count, off_t offset);
//
///*
// * Write a file to an MTD device
// * @param	path	file to write or a URL
// * @param	mtd	path to or partition name of MTD device 
// * @return	0 on success and errno on failure
// */
//int
//mtd_write(const char *path, const char *mtd)
//{
//	int mtd_fd = -1;
//	mtd_info_t mtd_info;
//	erase_info_t erase_info;
//
//	struct sysinfo info;
//	struct trx_header trx;
//	unsigned long crc;
//
//	FILE *fp;
//	char *buf = NULL;
//	long count, len, off;
//	int ret = -1;
//	unsigned char t[8];
//	/* Examine TRX header */
//	/* Modify by Terence 20030613 for F/W Header Check */	
//	if ((fp = fopen(path, "r")))
//	{
//
//#if defined(MOTOROLA_SERIES_CHECK_FIRMWARE)
//		if( !strcmp( mtd, "linux" ) )
//		{
//				safe_fread(t, 1, 8, fp);
//				//printf("fread count is \n");
//				//printf("%02X %02X %02X %02X %02X %02X %02X %02X \n",t[0],t[1],t[2],t[3],t[4],t[5],t[6],t[7]);
//		
//				if(memcmp( t + 4, VDID, 4 ))
//				{
//					fprintf(stderr, "Model Number  not match.\n" );
//					goto fail ;
//				}
//				else
//				{
//					fprintf(stderr, "Model Number match.\n" );
//				}
//		}
//		
//		count = safe_fread(&trx, 1, sizeof(struct trx_header), fp);
//#else
//		count = safe_fread(&trx, 1, sizeof(struct trx_header), fp);
//#endif   
//    //printf("mtd.c: trx.magic is %x\n",trx.magic);
//    //printf("mtd.c: trx.len is %x\n",trx.len);
//		//printf("mtd.c: count is %d\n",count);
//		//fprintf(stderr, "count=<%d>\n", count);		
//
//	}	
//	else
//		count = http_get(path, (char *) &trx, sizeof(struct trx_header), 0);
//	if (count < sizeof(struct trx_header)) {
//		fprintf(stderr, "%s: Truncated file\n", path);
//		goto fail;
//	}
//	if (trx.magic != TRX_MAGIC ||
//	    trx.len > TRX_MAX_LEN ||
//	    trx.len < sizeof(struct trx_header)) {
//		fprintf(stderr, "%s: Bad header\n", path);
//		goto fail;
//	}
//
//	/* Open MTD device and get sector size */
//
//	if ((mtd_fd = mtd_open(mtd, O_RDWR)) < 0 ||
//	    ioctl(mtd_fd, MEMGETINFO, &mtd_info) != 0 ||
//	    mtd_info.erasesize < sizeof(struct trx_header)) {
//		perror(mtd);
//		goto fail;
//	}
//
//	/* See if we have enough memory to store the whole file */
//	sysinfo(&info);
//	if ((info.freeram + info.bufferram) >= trx.len)
//		erase_info.length = ROUNDUP(trx.len, mtd_info.erasesize);
//	else
//		erase_info.length = mtd_info.erasesize;
//
//	/* Allocate temporary buffer */
//	if (!(buf = malloc(erase_info.length))) {
//		perror("malloc");
//		goto fail;
//	}
//
//	/* Write file or URL to MTD device */
//	for (erase_info.start = 0; erase_info.start < trx.len; erase_info.start += count) {
//		len = MIN(erase_info.length, trx.len - erase_info.start);
//		if (erase_info.start == 0) {
//			count = off = sizeof(struct trx_header);
//			memcpy(buf, &trx, sizeof(struct trx_header));
//		} else
//			count = off = 0;
//		if (fp)
//		{
//			count += fread(&buf[off], 1, len - off, fp);
//			
//			/* Modify by Terence 20030613 for F/W Header Check */
//
//#ifdef MOTOROLA_SERIES_CHECK_FIRMWARE
//
//			if( !strcmp( mtd, "linux" ) )
//			{
//				//fprintf(stderr, "count1=<%d> <%d> <%d>\n", count, trx.len, len);					
//
//				count += 8 ;
//				//fprintf(stderr, "count2=<%d> <%d> <%d>\n", count, trx.len, len);				
//			}
//
//#endif			
//
//				//fprintf(stderr, "count3=<%d>\n", count);
//
//				//fprintf(stderr, "buf[off]=<%2X>\n", buf[off]);
//		}
//		else
//			count += http_get(path, &buf[off], len - off, erase_info.start + off);
//		if (count < len) {
//			fprintf(stderr, "%s: Truncated file\n", path);
//			goto fail;
//		}
//		/* Check CRC before writing if possible */
//		if (count == trx.len) {
//			unsigned int moto_crc = 0;
//			fprintf(stderr, "Check CRC\n");
//			
//			crc = crc32(&buf[OFFSET(&trx, flag_version)],
//				    trx.len - OFFSET(&trx, flag_version),
//				    CRC32_INIT_VALUE);
//			if (crc != trx.crc32) {
//				fprintf(stderr, "%s: Bad CRC\n", path);
//				goto fail;
//			}
//			//fprintf(stderr, "Finish check CRC : %08X\n", crc);
//			
//			//printf("Check : Motorola crc : %02X%02X%02X%02X\n", t[0], t[1], t[2], t[3]);
//			
//			moto_crc = crc32(t + 4, 2, CRC32_INIT_VALUE);
//			moto_crc = crc32(t + 6, 2, moto_crc);
//			moto_crc = crc32(buf, trx.len, moto_crc);
//
//			if(!memcmp(t + 3, (char *)&moto_crc, 1) && !memcmp(t + 2, (char *)&moto_crc + 1, 1) && !memcmp(t + 1, (char *)&moto_crc + 2, 1) && !memcmp(t + 0, (char *)&moto_crc + 3, 1))
//			{
//				fprintf(stderr, "Motorola CRC correct...\n");
//			}
//			else
//			{
//				fprintf(stderr, "Motorola CRC incorrect...\n");
//				goto fail;
//			}
//		}
//		
//		//fprintf(stderr, "Write to mtd\n");
//		/* Do it */
//		if( ioctl(mtd_fd, MEMERASE, &erase_info) != 0 )
//		{
//			perror(mtd);
//			goto fail;
//		}			
//		if( !strcmp( mtd, "pmon" ) )
//		{
//			buf += 28 ;
//			if( write(mtd_fd, buf, count-28) != (count-28) )
//			{
//				perror(mtd);
//				goto fail;
//			}			
//		}
//		else
//		{
//			if( write(mtd_fd, buf, count) != count ) 
//			{
//				perror(mtd);
//				goto fail;
//			}			
//		}		
//		/* Do it */	
///*		if (ioctl(mtd_fd, MEMERASE, &erase_info) != 0 ||
//		    write(mtd_fd, buf, count) != count) {
//			perror(mtd);
//			goto fail;
//		}
//*/		
//	}
//	
//	//fprintf(stderr, "Finish write to mtd\n");
//
//	//fprintf(stderr, "Check CRC after writing\n");
//
//	/* Check CRC after writing */
//	lseek(mtd_fd, OFFSET(&trx, flag_version), SEEK_SET);
//	crc = CRC32_INIT_VALUE;
//	for (count = OFFSET(&trx, flag_version); count < trx.len;  count += erase_info.length) {
//		if (read(mtd_fd, buf, MIN(erase_info.length, trx.len - count)) < 0) {
//			perror(mtd);
//			goto fail;
//		}
//		crc = crc32(buf, MIN(erase_info.length, trx.len - count), crc);
//	}
//	//fprintf(stderr, "Finish check CRC after writing\n");	
//	//printf("CRC : %08X\n", crc);
//	
//	if (crc != trx.crc32) {
//		fprintf(stderr, "%s: Bad CRC\n", path);
//		goto fail;
//	} else
//		fprintf(stderr, "%s: CRC OK\n", mtd);
//
//	ret = 0;
// fail:
//	if (buf)
//		free(buf);
//	if (mtd_fd >= 0)
//		close(mtd_fd);
//	if (fp)
//		fclose(fp);
//	return ret;
//}
